import { equiv as _equiv } from "@thi.ng/equiv";
let $cachedFP;
let $cachedPath;
let $cachedEPC = [];
let $cachedPathPos = [];
const __cachedFP = (size) => $cachedFP && $cachedFP.length >= size ? $cachedFP : $cachedFP = new Int32Array(size);
const __cachedPath = (size) => $cachedPath && $cachedPath.length >= size ? $cachedPath : $cachedPath = new Int32Array(size);
const clearCache = () => {
  $cachedFP = $cachedPath = void 0;
  $cachedEPC.length = $cachedPathPos.length = 0;
};
const __simpleDiff = (state, src, key, logDir, mode) => {
  const n = src.length;
  const linear = state.linear;
  state.distance = n;
  if (mode !== "only-distance") {
    for (let i = 0, j = 0; i < n; i++, j += 3) {
      linear[j] = logDir;
      linear[j + 1] = i;
      linear[j + 2] = src[i];
    }
    if (mode === "full") {
      const _state = state[key];
      for (let i = 0; i < n; i++) {
        _state[i] = src[i];
      }
    }
  }
  return state;
};
const diffArray = (a, b, mode = "full", equiv = _equiv) => {
  const state = {
    distance: 0,
    adds: {},
    dels: {},
    const: {},
    linear: []
  };
  if (a === b || a == null && b == null) {
    return state;
  } else if (a == null || a.length === 0) {
    return __simpleDiff(state, b, "adds", 1, mode);
  } else if (b == null || b.length === 0) {
    return __simpleDiff(state, a, "dels", -1, mode);
  }
  const reverse = a.length >= b.length;
  let _a, _b, na, nb;
  if (reverse) {
    _a = b;
    _b = a;
  } else {
    _a = a;
    _b = b;
  }
  na = _a.length;
  nb = _b.length;
  const offset = na + 1;
  const delta = nb - na;
  const doff = delta + offset;
  const size = na + nb + 3;
  const path = __cachedPath(size).fill(-1, 0, size);
  const fp = __cachedFP(size).fill(-1, 0, size);
  const epc = $cachedEPC;
  const pathPos = $cachedPathPos;
  const snake = (k2, p2, pp) => {
    let r, y;
    if (p2 > pp) {
      r = path[k2 - 1];
      y = p2;
    } else {
      r = path[k2 + 1];
      y = pp;
    }
    let x = y - (k2 - offset);
    while (x < na && y < nb && equiv(_a[x], _b[y])) {
      x++;
      y++;
    }
    path[k2] = pathPos.length / 3;
    pathPos.push(x, y, r);
    fp[k2] = y;
  };
  let p = -1;
  let k;
  do {
    p++;
    for (k = -p + offset; k < doff; k++) {
      snake(k, fp[k - 1] + 1, fp[k + 1]);
    }
    for (k = doff + p; k > doff; k--) {
      snake(k, fp[k - 1] + 1, fp[k + 1]);
    }
    snake(doff, fp[doff - 1] + 1, fp[doff + 1]);
  } while (fp[doff] !== nb);
  state.distance = delta + 2 * p;
  if (mode !== "only-distance") {
    p = path[doff] * 3;
    while (p >= 0) {
      epc.push(p);
      p = pathPos[p + 2] * 3;
    }
    if (mode === "full") {
      __buildFullLog(epc, pathPos, state, _a, _b, reverse);
    } else {
      __buildLinearLog(
        epc,
        pathPos,
        state,
        _a,
        _b,
        reverse,
        mode === "only-distance-linear"
      );
    }
  }
  epc.length = 0;
  pathPos.length = 0;
  return state;
};
const __buildFullLog = (epc, pathPos, state, a, b, reverse) => {
  const linear = state.linear;
  const _const = state.const;
  let i = epc.length;
  let px = 0;
  let py = 0;
  let adds;
  let dels;
  let aID;
  let dID;
  if (reverse) {
    adds = state.dels;
    dels = state.adds;
    aID = -1;
    dID = 1;
  } else {
    adds = state.adds;
    dels = state.dels;
    aID = 1;
    dID = -1;
  }
  for (; i-- > 0; ) {
    const e = epc[i];
    const ppx = pathPos[e];
    const ppy = pathPos[e + 1];
    const d = ppy - ppx;
    while (px < ppx || py < ppy) {
      const dp = py - px;
      if (d > dp) {
        linear.push(aID, py, adds[py] = b[py]);
        py++;
      } else if (d < dp) {
        linear.push(dID, px, dels[px] = a[px]);
        px++;
      } else {
        linear.push(0, px, _const[px] = a[px]);
        px++;
        py++;
      }
    }
  }
};
const __buildLinearLog = (epc, pathPos, state, a, b, reverse, inclConst) => {
  const linear = state.linear;
  const aID = reverse ? -1 : 1;
  const dID = reverse ? 1 : -1;
  let i = epc.length, px = 0, py = 0;
  for (; i-- > 0; ) {
    const e = epc[i];
    const ppx = pathPos[e];
    const ppy = pathPos[e + 1];
    const d = ppy - ppx;
    while (px < ppx || py < ppy) {
      const dp = py - px;
      if (d > dp) {
        linear.push(aID, py, b[py]);
        py++;
      } else if (d < dp) {
        linear.push(dID, px, a[px]);
        px++;
      } else {
        inclConst && linear.push(0, px, a[px]);
        px++;
        py++;
      }
    }
  }
};
export {
  clearCache,
  diffArray
};
